%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:FUNCTION-PRIMITIVES.SL
% Title:        Primitives used by PUTD/GETD and EVAL/APPLY
% Author:       Eric Benson, Steve Lowder
% Created:      23 August 1981
% Modified:     29-Aug-84 10:08:36 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:  PL:SYS-CONSTS.B
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 17-Jul-84 14:00 (Brian Beach)
%  Minor cleanup.  File is now portable, so changed directory in file
%  name to PNK:.
% 27-Mar-84 13:56:06 (Jim Ambras/CRC)
%  Changed use of idloc to id2int.  Added compiletime load of sys-consts.
% 02-Mar-84 09:19:52 (Jim Ambras/CRC)
%  Corrected file header.
% 13-Dec-83 12:27:47 (Tim Tillson)
%  Modified for the D-register model; plant routines moved to LAP.SL
% 02-Dec-83 10:14:03 (Jim Ambras)
%   Translated from Rlisp to Lisp.
% 15-Jul-83 Steve Lowder
%  Changed three *WTimes2 expressions for (WConst 6) to adds because *WTimes2
%  is no longer a cmacro.
% 1-Jul-83 Steve Lowder
%  Changed expressions (displacement (reg 1) 0) to (indirect (reg 1)) to
%  improve code generation.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%
% Every ID has a "function cell".  It does not necessarily contain a legal
% Lisp item, and therefore should not be accessed directly by Lisp functions.
% In this implementation the function cell contains an instruction to be
% executed.  There are 3 possibilites for this instruction, for which the
% following predicates and updating functions exist:
%
%    FUnBoundP(ID)          -- the function is not defined
%    FLambdaLinkP(ID)       -- the function is interpreted
%    FCodeP(ID)             -- the function is compiled
%
%    MakeFUnBound(ID)       -- undefine the function
%    MakeFLambdaLink(ID)    -- specify that the function is interpreted
%    MakeFCode(ID, CodePtr) -- specify that the function is compiled,
%                              and that the code resides at the address
%                              associated with CodePtr
%
%    GetFCodePointer(ID)    -- returns the contents of the function cell as a
%                              code pointer.
%
% These functions currently check that they have proper arguments, but this may
% change since they are only used by functions that have checked them already.
%
 
(compiletime (load sys-consts))
 
(fluid '(lambda-link-address undefined-function-address))
 
(setq lambda-link-address
  (wplus2 symfnc (idinf 'compiledcallinginterpreted)))
 
(setq undefined-function-address
  (wplus2 symfnc (idinf 'undefinedfunction)))
 
(de funboundp (u)
  % does U not have a function defn?
  (if (idp u)
    (weq (symfnc (idinf u)) undefined-function-address)
    (noniderror u 'funboundp)))
 
(de flambdalinkp (u)
  % is U an interpreted function?
  (if (idp u)
    (weq (symfnc (idinf u)) lambda-link-address)
    (noniderror u 'flambdalinkp)))
 
(de fcodep (u)
  % is U a compiled function?
  (if (idp u)
    (and (wneq (symfnc (idinf u)) undefined-function-address)
       (wneq (symfnc (idinf u)) lambda-link-address))
    (noniderror u 'fcodep)))
 
(de makefunbound (u)
  % Make U an undefined function
  (if (idp u)
    (progn (plantunbound (idinf u))
           nil)
    (noniderror u 'makefunbound)))
 
(de makeflambdalink (u bit)
  % Make U an interpreted function
  (if (idp u)
    (progn (plantlambdalink (idinf u) bit)
           nil)
    (noniderror u 'makeflambdalink)))
 
(de makefcode (u codeptr bit)
  % Make U a compiled function
  (when (idp u)
    (if (codep codeptr)
      (progn (plantcodepointer (idinf u) (inf codeptr) bit)
             nil)
      (noniderror u 'makefcode))))
 
(de getfcodepointer (u)
  % Get code pointer for U
  (if (idp u)
    (mkcode (symfnc (idinf u)))
    (noniderror u 'getfcodepointer)))
 
 
% P-FUNCTION-PRIMITIVES
%
% Author:      M. L. Griss
%             Symbolic Computation Group
%              Computer Science Dept.
%              University of Utah
% Date:        21 October 1982
% Copyright (c) 1982 University of Utah
%
% 15-May-85 Kessler - Hand coded fastapply for the CRAY.  Should be
%  significantly faster.  IN the reduce test, fastapply is called 4906 times.
% 13-May-85 Kessler - Removed Linkregs
% Based on P20:Function-Primitives.Red
%  <PSL.TESTS>P-FUNCTION-PRIMITIVES.RED.4,  2-Mar-83 11:46:30, Edit by KESSLER
%  Put in Dealloc's before jump and jcall (search rrk)
 
% Every ID has a "function cell".  It does not necessarily contain a legal
% Lisp item, and therefore should not be accessed directly by Lisp functions.
% In this implementation the function cell contains an instruction to be
% executed.  There are 3 possibilites for this instruction, for which the
% following predicates and updating functions exist:
%
%      FUnBoundP(ID) -- the function is not defined
%      FLambdaLinkP(ID) -- the function is interpreted
%      FCodeP(ID) -- the function is compiled
%
%      MakeFUnBound(ID) -- undefine the function
%      MakeFLambdaLink(ID) -- specify that the function is interpreted
%      MakeFCode(ID, CodePtr) -- specify that the function is compiled,
%                           and that the code resides at the address
%                           associated with CodePtr
%
%      GetFCodePointer(ID) -- returns the contents of the function cell as a
%                        code pointer
%
% See the templates in XXX-ASM.RED:
%
%       DefinedFunctionCellFormat*
%      UndefinedFunctionCellFormat*
 
 
% These functions currently check that they have proper arguments,
% but this may change since they are only used by functions that
% have checked them already.
 
% Note that on some machines, SYMFNC(x) is entire SYMFNC cell.
%           on others it points into the cell, at the "address" part.
%
% Fairly Portable versions, based on assumption that
%      Starts with OPCODE, probably *JCALL
%      *Jcall SymfncBase UndefinedFunction   in ShouldBeUndefined cell
 
% Needs the machine-dependent procedures in XXX-HEADER:
 
%    %Store-JCALL(CodeAddress,StoreAddress)
%        to Create a *Jcall(CodeAddress) at StoreAddress
 
%    %Copy-Function-Cell(From,to)
%        to copy appropriate # words or bytes of Function cell
(Fluid '(CodePtr* CodeForm* CodeNarg* CodeArgs))
 
(LAP '((*entry CodePrimitive expr 15)
%      Takes the code pointer stored in the fluid variable CodePtr*
%      and jumps to its address, without disturbing any of the argument
%      registers.  This can be flagged 'InternalFunction for compilation
%      before this file is compiled or done as an 'OpenCode and 'ExitOpenCode
%      property for the compiler.
      (*ALLOC 0)
      (*MOVE ($Fluid CodePtr*) (reg t1))
      (*FIELD (reg t1) (reg t1)     % get CodeINF
              (WConst InfStartingBit) (WConst InfBitLength))
      (*Dealloc 0)
      (*JUMP (memory (reg t1) (Wconst 0)))
      (*EXIT 0)
)    )
 
 
(LAP '((*entry CompiledCallingInterpreted expr 15)
%      Called by some convention from the function cell of an ID which
%      has an interpreted function definition.  It should store the
%       Linkreg into
%       the fluid variable CodeForm* without disturbing the argument
%      registers
%
%
      (*ALLOC 0)
      (*CALL SaveRegisters)      % *CALL to avoid resetting LinkInfo
      (*move (reg NargReg) (Reg S6))
      (*field (Reg s6) (reg s6) (wconst &LinkShiftPos&)
          (wconst &LinkShiftBits&))
      (*move (reg s6) ($Fluid CodeNarg*))
      (*move (reg NargReg) (Reg S7))
      (*wshift (reg S7) (minus (wconst &LinkShiftBits&)))
      (*move (reg s7) ($Fluid CodeForm*))
      (*Dealloc 0)
      (*JCALL CompiledCallingInterpretedAux)
      (*Exit 0)
)    )
 
 
(LAP '((*entry FastApply expr 0)
%      Called with a functional form in (reg t1) and argument registers
%      loaded.  If it is a code pointer or an ID, the function address
%      associated with either should be jumped to.  If it is anything else
%      except a lambda form, an error should be signaled.  If it is a lambda
%      form, store (reg t1) in the fluid variable CodeForm* and
%      (*JCALL FastLambdaApply)
%      (FastLambdaApply may be flagged 'InternalFunction).
      (*ALLOC 0)
      (MOVE (reg s6) (reg t1))
      (*FIELD (reg s6) (reg s6) (wconst tagstartingbit)
            (wconst tagbitlength))
      (*JUMPNOTEQ (label notanid) (reg s6) (wconst id-tag))
      (move (reg s7) (reg nargreg))
      (move (reg s6) 15)
      (and (reg s7) (reg s7) (reg s6))
      (move (reg s6) (reg t1))
      (*wshift (reg s6) (wconst TagBitLength))
      (*wshift (reg s6) (minus (difference (wconst tagBitlength)
                        (wconst &LinkShiftBits&))))
      (or (reg s7) (reg s7) (reg s6))
      (move (reg Nargreg) (reg s7))
      (*Dealloc 0)
      (move (reg s6) (reg t1))
      (*move ($fluid symfnc) (reg a6))
      (*move (reg s6) (reg a7))
      (add (reg a6) (reg a6) (reg a7))
      (load (reg s6) (indexed (reg a6) 0))
      (*wshift (reg s6) -32)
      (move (reg a6) (reg s6))
      (move (reg b02) (reg a6))
      (j (reg b02))
NotAnID  % S6 has the tag (risky since jump must not wreck s6)
      (*JUMPNOTEQ (Label NotACodePointer) (reg s6) (WConst CODE-tag))
      (*Dealloc 0)
      (move (reg s6) (reg t1))
      (lshift (reg s6) (reg s6) 2)
      (move (reg a6) (reg s6))
      (move (reg b02) (reg a6))
      (j (reg b02))
NotACodePointer  % S6 still has the tag.
      (*JUMPNOTEQ (Label IllegalFunctionalForm) (reg s6)
                          (wconst pair-tag))
      (*MOVE (MEMORY (reg t1) (WConst 0)) (reg s6))
                              % CAR with pair already untagged
      (*JUMPNOTEQ (Label IllegalFunctionalForm) (reg s6) (QUOTE LAMBDA))
      (*Dealloc 0)
      (move (reg s6) (reg t1))
      (*MOVE (REG S6) ($FLUID Codeform*))
      (move (reg a6) (reg s6))   % Strip off any tag.
      (move (reg s6) (reg a6))
      (move (reg t1) (reg s6))   % Ok, T1 has the guy to apply.
      (*JCALL FastLambdaApply)                % CodeForm*
                                    % Already Loaded
IllegalFunctionalForm
      (*JCALL ApplyError)
)    )
 
(lap '((*Entry ApplyError Expr 0)
      (*ALLOC 0)
      (*MOVE (QUOTE "Illegal functional form in Apply") (reg 1))
      (*MOVE (reg t1) (reg 2))
      (*CALL List2)
% rrk 03/03/83
      (*Dealloc 0)
      (*JCALL StdError)
)     )
 
(de SaveRegisters (A1 A2 A3 A4 A5 A6 A7 A8 A9 A10 A11 A12 A13 A14 A15)
   (progn
    (wputv CodeArgs 14   A15)
    (wputv CodeArgs 13   A14)
    (wputv CodeArgs 12   A13)
    (wputv CodeArgs 11   A12)
    (wputv CodeArgs 10   A11)
    (wputv CodeArgs 9   A10)
    (wputv CodeArgs 8   A9)
    (wputv CodeArgs 7   A8)
    (wputv CodeArgs 6   A7)
    (wputv CodeArgs 5   A6)
    (wputv CodeArgs 4   A5)
    (wputv CodeArgs 3   A4)
    (wputv CodeArgs 2   A3)
    (wputv CodeArgs 1   A2)
    (wputv CodeArgs 0   A1)
)  )
 
(LAP '((*ENTRY UndefinedFunctionAux expr 0)
%      Called by some convention from the function cell of an ID (probably
%      the same as CompiledCallingInterpreted) for an undefined function.
%      Should call Error with the ID as part of the error message.
      (*ALLOC 0)
      (*CALL SaveRegisters)    % *CALL so as not to change LinkInfo
                               % Was stored in UndefnCode* UndefnNarg*
      (*CALL UndefinedFunctionAuxAux)
      (*Dealloc 0)
      (*move 4711 (reg 1))
      (*move (quote "Undefined function called") (reg 2))
      (*JCALL error)
)    )
(de error (x) (kernel-fatal-error x)) % until error is present
